# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2024-2025 Mira Geoscience Ltd.                                '
#                                                                              '
#  This file is part of curve-apps package.                                    '
#                                                                              '
#  curve-apps is distributed under the terms and conditions of the MIT License '
#  (see LICENSE file at the root of this source code package).                 '
#                                                                              '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

from __future__ import annotations

from pathlib import Path
from typing import ClassVar

from geoapps_utils.driver.data import BaseData
from geoh5py.data import Data, ReferencedData
from geoh5py.groups import UIJsonGroup
from geoh5py.objects import Curve, Points
from pydantic import BaseModel, ConfigDict

from curve_apps import assets_path


class TrendLineSourceParameters(BaseModel):
    """
    Source parameters expected by the ui.json file format.

    :param entity: A Curve or Points source object.
    :param data: Data values to find edges on.
    :param parts: Optional parts to connect.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    entity: Curve | Points
    data: ReferencedData | None = None
    parts: Data | None = None


class TrendLineDetectionParameters(BaseModel):
    """
    Detection parameters expected by the ui.json file format.

    :param azimuth: Azimuth of the path.
    :param azimuth_tol: Tolerance for the azimuth of the path.
    :param damping: Damping factor between [0, 1] for the path roughness.
    :param min_edges: Minimum number of points in a curve.
    :param max_distance: Maximum distance between points in a curve.
    """

    azimuth: float | None = None
    azimuth_tol: float | None = None
    damping: float = 0
    min_edges: int = 1
    max_distance: float | None = None


class TrendLineOutputParameters(BaseModel):
    """
    Output parameters.

    :param export_as: Name of the output entity.
    :param out_group: Name of the output group.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    export_as: str | None = "trend_lines"
    out_group: UIJsonGroup | None = None


class TrendLineParameters(BaseData):
    """
    Trend lines parameters for use with `trend_lines.driver`.

    :param source: Source data parameters.
    :param detection: Trend line detection parameters.
    :param output: Trend line output parameters.
    """

    name: ClassVar[str] = "trend_lines"
    default_ui_json: ClassVar[Path] = assets_path() / "uijson/trend_lines.ui.json"
    title: ClassVar[str] = "Trend Lines Detection"
    run_command: ClassVar[str] = "curve_apps.trend_line_detection.driver"

    conda_environment: str = "curve_apps"
    source: TrendLineSourceParameters
    detection: TrendLineDetectionParameters
    output: TrendLineOutputParameters = TrendLineOutputParameters()



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2024-2025 Mira Geoscience Ltd.                                '
#                                                                              '
#  This file is part of curve-apps package.                                    '
#                                                                              '
#  curve-apps is distributed under the terms and conditions of the MIT License '
#  (see LICENSE file at the root of this source code package).                 '
#                                                                              '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import logging
import sys

import numpy as np
from geoapps_utils.utils.locations import (
    get_overlapping_limits,
    map_indices_to_coordinates,
)
from geoh5py.data import FloatData
from geoh5py.objects import Curve, Grid2D
from geoh5py.ui_json import InputFile, utils
from scipy.spatial import cKDTree
from skimage.feature import canny  # pylint: disable=no-name-in-module
from skimage.transform import (  # pylint: disable=no-name-in-module
    probabilistic_hough_line,
)

from curve_apps.driver import BaseCurveDriver
from curve_apps.edges.params import EdgeDetectionParameters, EdgeParameters


logger = logging.getLogger(__name__)


class EdgesDriver(BaseCurveDriver):
    """
    Driver for the edge detection application.

    :param parameters: Application parameters.
    """

    _parameter_class = EdgeParameters

    def __init__(self, parameters: EdgeParameters | InputFile):
        super().__init__(parameters)

    def make_curve(self):
        """
        Make curve object from edges detected in source data.

        The application relies on the Canny and Hough transforms from the
        Scikit-Image library.

        """
        with utils.fetch_active_workspace(self.workspace, mode="r+") as workspace:
            logging.info("Generated edges ...")
            canny_grid = EdgesDriver.get_canny_edges(
                self.params.source.objects,
                self.params.source.data,
                self.params.detection,
            )
            vertices, cells = EdgesDriver.get_edges(
                self.params.source.objects,
                canny_grid,
                self.params.detection,
            )

            if vertices is None or cells is None:
                return None

            self.params.source.objects.add_data(
                {"canny filter": {"values": canny_grid.flatten(order="F")}}
            )
            curve = Curve.create(
                workspace=workspace,
                name=self.params.output.export_as,
                vertices=vertices,
                cells=cells,
                parent=self.out_group,
            )

            # Compute positive angle from North
            # TODO: Move to geoapps-utils
            delta = np.c_[
                vertices[cells[:, 1], 0] - vertices[cells[:, 0], 0],
                vertices[cells[:, 1], 1] - vertices[cells[:, 0], 1],
            ]
            delta[delta[:, 0] < 0, :] *= -1
            amp = np.linalg.norm(delta, axis=1)
            orientation = np.arccos(delta[:, 1] / amp)

            # TODO: Assign values to vertices until better handling of cell data by GA
            vert_azimuth = np.zeros(curve.n_vertices) * np.nan
            vert_azimuth[cells.flatten()] = np.repeat(orientation, 2)
            curve.add_data(
                {
                    "azimuth": {"values": np.degrees(vert_azimuth)},
                }
            )

            vert_lengths = np.zeros(curve.n_vertices) * np.nan
            vert_lengths[cells.flatten()] = np.repeat(amp, 2)
            curve.add_data(
                {
                    "lengths": {"values": vert_lengths},
                }
            )

        return curve

    @staticmethod
    def get_canny_edges(
        grid: Grid2D, data: FloatData, detection: EdgeDetectionParameters
    ) -> np.ndarray:
        """
        Get edges from a grid.

        :param grid: Grid2D object.
        :param data: FloatData object.
        :param detection: Detection parameters.

        :returns: Edges from Canny transform.
        """
        if grid.shape is None or data.values is None:
            raise ValueError("Grid and data must be defined.")

        grid_data = data.values.reshape(grid.shape, order="F")

        if np.all(np.isnan(grid_data)):
            raise ValueError("No data to process.")

        # Find edges
        edges = canny(
            grid_data,
            sigma=detection.sigma,
            use_quantiles=True,
            mask=~np.isnan(grid_data),
            mode="reflect",
        )
        grid.add_data({"canny filter": {"values": edges.flatten(order="F")}})

        return edges

    @staticmethod
    def get_edges(
        grid: Grid2D,
        edges: np.ndarray,
        detection: EdgeDetectionParameters,
    ) -> tuple[np.ndarray, np.ndarray] | tuple[None, None]:
        """
        Find edges in gridded data.

        :params grid: A Grid2D object.
        :params edges: Edges representation of the grid from Canny transform.
        :params detection: Detection parameters.

        :returns : n x 3 array. Vertices of edges.
        :returns : n x 2 float array. Cells of edges.
        """
        # Find lines
        indices = EdgesDriver.get_line_indices(
            edges,
            detection.line_length,
            detection.line_gap,
            detection.threshold,
            detection.window_size,
        )

        if len(indices) == 0:
            return None, None

        pixel_coordinates = np.vstack(indices)

        if detection.merge_length is not None:
            vertices = map_indices_to_coordinates(grid, pixel_coordinates)
            tree = cKDTree(vertices)
            merge = tree.query_pairs(detection.merge_length, output_type="ndarray")
            pixel_coordinates[merge[:, 0], :] = pixel_coordinates[merge[:, 1], :]

        u_ind, r_ind = np.unique(pixel_coordinates, axis=0, return_inverse=True)
        vertices = map_indices_to_coordinates(grid, u_ind)

        cells = r_ind.reshape((-1, 2))
        return vertices, cells

    @staticmethod
    def get_line_indices(
        canny_image: np.ndarray,
        line_length: int = 1,
        line_gap: int = 1,
        threshold: int = 1,
        window_size: int | None = None,
    ) -> list:
        """
        Get indices forming lines on a canny image.

        The process is done over tiles of square size. The tiles overlap by 25%.

        :param canny_image: Edges.
        :param line_length: Minimum accepted pixel length of detected lines. (Hough)
        :param line_gap: Maximum gap between pixels to still form a line. (Hough)
        :param threshold: Value threshold. (Hough)
        :param window_size: Size of the window to search for lines.

        :returns: List of indices.
        """
        # Cycle through tiles of square size
        width = np.min(canny_image.shape)
        if window_size is not None:
            width = np.min([window_size, width])

        x_limits = get_overlapping_limits(canny_image.shape[0], width)
        y_limits = get_overlapping_limits(canny_image.shape[1], width)

        # TODO: Loop in parallel
        indices = []
        for x_lim in x_limits:
            for y_lim in y_limits:
                lines = probabilistic_hough_line(
                    canny_image[x_lim[0] : x_lim[1], y_lim[0] : y_lim[1]],
                    line_length=line_length,
                    threshold=threshold,
                    line_gap=line_gap,
                    rng=0,
                )

                if np.any(lines):
                    # Add the limits of the tile to the indices
                    lines = np.vstack(lines)[:, ::-1] + np.c_[x_lim[0], y_lim[0]]
                    indices.append(lines)

        return indices


if __name__ == "__main__":
    file = sys.argv[1]
    ifile = InputFile.read_ui_json(file)

    driver = EdgesDriver(ifile)
    driver.run()



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2024-2025 Mira Geoscience Ltd.                                '
#                                                                              '
#  This file is part of curve-apps package.                                    '
#                                                                              '
#  curve-apps is distributed under the terms and conditions of the MIT License '
#  (see LICENSE file at the root of this source code package).                 '
#                                                                              '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

# pylint: disable=duplicate-code

from __future__ import annotations

import logging
import sys

import numpy as np
from geoh5py.objects import Curve
from geoh5py.ui_json import InputFile, utils
from tqdm import tqdm

from curve_apps.driver import BaseCurveDriver
from curve_apps.trend_lines.params import TrendLineParameters
from curve_apps.utils import find_curves


logger = logging.getLogger(__name__)


class TrendLinesDriver(BaseCurveDriver):
    """
    Driver for the detection of trend curves across an object parts.

    :param parameters: Application parameters.
    """

    _parameter_class = TrendLineParameters

    def __init__(self, parameters: TrendLineParameters | InputFile):
        super().__init__(parameters)

    def make_curve(self):
        """Make curve object from trend lines detected in source data."""

        with utils.fetch_active_workspace(self.workspace, mode="r+") as workspace:
            logging.info("Generating trend lines ...")
            vertices, cells, labels = self.get_connections()

            if cells is None:
                logger.info("No connections found.")
                return None

            curve = Curve.create(
                workspace=workspace,
                name=self.params.output.export_as,
                vertices=vertices,
                cells=cells,
                parent=self.out_group,
            )

            if curve is not None and self.params.source.data is not None:
                curve.add_data(
                    {
                        self.params.source.data.name: {
                            "values": labels,
                            "entity_type": self.params.source.data.entity_type,
                            "association": "VERTEX",
                        }
                    }
                )

        return curve

    def get_connections(self) -> tuple:
        """
        Find connections between entity parts.

        :params entity: A Curve object with parts.
        :params data: Input referenced data.
        :params detection: Detection parameters.

        :returns : n x 3 array. Vertices of connecting lines.
        :returns : n x 2 float array. Cells of edges.
        :returns : n x 1 array. Labels of vertices.
        """
        path_list = []
        out_labels = np.zeros_like(self.labels).astype("int32")

        for value in tqdm(np.unique(self.labels), desc="Looping over data labels"):
            if value == 0:
                continue

            ind = np.where(self.labels == value)[0]

            if len(ind) < 2:
                continue

            segments = find_curves(
                self.vertices[ind, :2],
                self.parts[ind],
                self.params.detection,
            )

            if any(segments):
                path_list += ind[np.vstack(segments)].tolist()
                out_labels[ind] = value

        if any(path_list):
            path = np.vstack(path_list)

            # Truncate vertices and renumber
            verts_bool = np.zeros(self.vertices.shape[0], dtype=bool)

            uni_ind = np.unique(path.flatten())

            verts_bool[uni_ind] = True
            new_indices = np.ones_like(verts_bool, dtype="int32")
            new_indices[verts_bool] = np.arange(uni_ind.shape[0])
            path = new_indices[path]

            return (
                self.vertices[uni_ind, :],
                path,
                out_labels[uni_ind],
            )

        return self.vertices, None, out_labels

    @property
    def vertices(self) -> np.ndarray:
        """
        Get vertices from entity.
        """
        entity = self.params.source.entity
        if entity.vertices is None:
            raise ValueError("Curve must have vertices to find connections.")

        return entity.vertices

    @property
    def parts(self) -> np.ndarray:
        """
        Get parts from entity or data.
        """
        entity = self.params.source.entity
        if self.params.source.parts is not None:
            return self.params.source.parts.values

        if isinstance(entity, Curve) and entity.parts is not None:
            return entity.parts

        return np.arange(self.vertices.shape[0]).astype("int32")

    @property
    def labels(self) -> np.ndarray:
        """
        Get labels from data.
        """
        data = self.params.source.data
        if data is not None and data.values is not None:
            values = data.values
        else:
            values = np.ones_like(self.parts)

        return values


if __name__ == "__main__":
    file = sys.argv[1]
    # file = r"C:\Users\dominiquef\Desktop\parts.ui.json"
    ifile = InputFile.read_ui_json(file)

    driver = TrendLinesDriver(ifile)
    driver.run()



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2024-2025 Mira Geoscience Ltd.                                '
#                                                                              '
#  This file is part of curve-apps package.                                    '
#                                                                              '
#  curve-apps is distributed under the terms and conditions of the MIT License '
#  (see LICENSE file at the root of this source code package).                 '
#                                                                              '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

from __future__ import annotations

import logging
from pathlib import Path


__version__ = "0.2.0-rc.1"
logging.basicConfig(level=logging.INFO)


def assets_path() -> Path:
    """Return the path to the assets folder."""

    parent = Path(__file__).parent
    folder_name = f"{parent.name}-assets"
    assets_folder = parent.parent / folder_name
    if not assets_folder.is_dir():
        raise RuntimeError(f"Assets folder not found: {assets_folder}")

    return assets_folder



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2024-2025 Mira Geoscience Ltd.                                '
#                                                                              '
#  This file is part of curve-apps package.                                    '
#                                                                              '
#  curve-apps is distributed under the terms and conditions of the MIT License '
#  (see LICENSE file at the root of this source code package).                 '
#                                                                              '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

from __future__ import annotations

from pathlib import Path
from typing import ClassVar

import numpy as np
from geoapps_utils.driver.data import BaseData
from geoh5py.data import Data
from geoh5py.groups import UIJsonGroup
from geoh5py.objects import Curve, Grid2D, Points, Surface
from geoh5py.ui_json.utils import str2list
from pydantic import BaseModel, ConfigDict, field_validator

from curve_apps import assets_path


class ContourSourceParameters(BaseModel):
    """
    Source parameters providing input data to the driver.

    :param objects: A Grid2D, Points, Curve or Surface source object.
    :param data: Data values to contour.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    objects: Grid2D | Points | Curve | Surface
    data: Data


class ContourDetectionParameters(BaseModel):
    """
    Contour specification parameters.

    :param interval_min: Minimum value for contours.
    :param interval_max: Maximum value for contours.
    :param interval_spacing: Step size for contours.
    :param fixed_contours: String defining list of fixed contours.
    :param max_distance: Maximum distance for interpolation.
    :param resolution: Resolution of underlying grid.
    """

    interval_min: float | None = None
    interval_max: float | None = None
    interval_spacing: float | None = None
    fixed_contours: list[float] | None = None
    max_distance: float = 500.0
    resolution: float = 50.0

    @field_validator("fixed_contours", mode="before")
    @classmethod
    def fixed_contour_input_to_list_of_floats(cls, val):
        """Parse fixed contour string into list of floats."""

        if isinstance(val, list):
            if not all(isinstance(k, float) for k in val):
                raise ValueError("List of fixed contours must contain only floats.")
            fixed_contours = val

        elif isinstance(val, str):
            fixed_contours = str2list(val)

        else:
            raise ValueError(
                "Fixed contours must be a list of floats, "
                "a string containing comma separated numeric characters, "
                "or None."
            )

        return fixed_contours

    @property
    def has_intervals(self) -> bool:
        """True if interval min, max and spacing are defined."""

        has_min_max = None not in [self.interval_min, self.interval_max]
        has_spacing = self.interval_spacing not in [0, None]

        return has_min_max and has_spacing

    @property
    def intervals(self) -> list[float]:
        """Returns arange of requested contour intervals."""

        if self.has_intervals:
            intervals = np.arange(
                self.interval_min,
                self.interval_max + self.interval_spacing / 2,  # type: ignore
                self.interval_spacing,
            ).tolist()
        else:
            intervals = []

        return intervals

    @property
    def contours(self) -> list[float]:
        """
        Returns a list of requested contours merging interval and fixed values.
        """
        contours = self.intervals + (self.fixed_contours or [])
        contours.sort()
        return contours


class ContourOutputParameters(BaseModel):
    """
    Output parameters.

    :param z_value: Use data values for curve height (z) channel
    :param export_as: Name of the output entity.
    :param out_group: Name of the output group.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    z_value: bool = False
    export_as: str | None = "Contours"
    out_group: UIJsonGroup | None = None


class ContourParameters(BaseData):
    """
    Contour parameters for use with `contours.driver`.

    :param contours: Contouring parameters.
    :param source: Parameters for the source object and data.
    :param output: Output
    """

    name: ClassVar[str] = "contours"
    default_ui_json: ClassVar[Path] = assets_path() / "uijson/contours.ui.json"
    title: ClassVar[str] = "Contour Detection"
    run_command: ClassVar[str] = "curve_apps.contour_detection.driver"

    conda_environment: str = "curve_apps"
    source: ContourSourceParameters
    detection: ContourDetectionParameters
    output: ContourOutputParameters = ContourOutputParameters()



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2024-2025 Mira Geoscience Ltd.                                '
#                                                                              '
#  This file is part of curve-apps package.                                    '
#                                                                              '
#  curve-apps is distributed under the terms and conditions of the MIT License '
#  (see LICENSE file at the root of this source code package).                 '
#                                                                              '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

from __future__ import annotations

import re
from collections.abc import Callable

import numpy as np
from geoapps_utils.utils.numerical import weighted_average
from geoh5py.objects import Curve, Grid2D, ObjectBase, Points, Surface
from scipy.interpolate import LinearNDInterpolator, interp1d
from scipy.spatial import Delaunay

from curve_apps.contours.params import ContourDetectionParameters
from curve_apps.trend_lines.params import TrendLineDetectionParameters


def image_to_grid_coordinate_transfer(
    image: np.ndarray, grid: list[np.ndarray]
) -> Callable:
    """
    Returns a function to interpolate from image to grid coordinates.

    :param grid: list of x and y grids.
    """
    row = np.arange(image.shape[0])
    col = np.arange(image.shape[1])
    x_interp = interp1d(col, grid[0])
    y_interp = interp1d(row, grid[1])

    def interpolator(col, row):
        return np.c_[x_interp(col), y_interp(row)]

    return interpolator


def interp_to_grid(
    entity: ObjectBase, values: np.ndarray, resolution: float, max_distance: float
) -> tuple[list[np.ndarray], np.ndarray]:
    """
    Interpolate values into a regular grid based on entity locations.

    :param entity: Geoh5py object with locations data.
    :param values: Data to be interpolated to grid.
    :param resolution: Grid resolution
    :param max_distance: Maximum distance used in weighted average.
    """

    if entity.locations is None:
        raise ValueError("Entity must have locations.")

    grid = []
    for dim in np.arange(2):
        grid += [
            np.arange(
                entity.locations[:, dim].min(),
                entity.locations[:, dim].max() + resolution,
                resolution,
            )
        ]

    x, y = np.meshgrid(grid[0], grid[1])
    z = np.zeros_like(x)
    values = weighted_average(
        entity.locations,
        np.c_[x.flatten(), y.flatten(), z.flatten()],
        [values],
        threshold=resolution / 2.0,
        n=8,
        max_distance=max_distance,
    )
    values = values[0].reshape(x.shape)

    return grid, values


def set_vertices_height(vertices: np.ndarray, entity: ObjectBase):
    """
    Uses entity z values to add height column to an Nx2 vertices array.

    :param vertices: Nx2 array of vertices.
    :param entity: geoh5py entity with vertices property.

    returns: Nx3 array of vertices.
    """
    if isinstance(entity, Points | Curve | Surface):
        if entity.vertices is None:
            raise ValueError("Entity does not have vertices.")
        z_interp = LinearNDInterpolator(
            entity.vertices[:, :2],
            entity.vertices[:, 2],
            fill_value=np.mean(entity.vertices[:, 2]),
        )
        vertices = np.c_[vertices, z_interp(vertices)]
    elif isinstance(entity, Grid2D):
        vertices = np.c_[
            vertices,
            np.ones(vertices.shape[0]) * entity.origin["z"],
        ]

    return vertices


def get_contour_list(params: ContourDetectionParameters) -> list[float]:
    """
    Compute contours requested by input parameters.

    :returns: Corresponding list of values in float format.
    """

    if (
        None not in [params.interval_min, params.interval_max, params.interval_spacing]
        and params.interval_spacing != 0
    ):
        interval_contours = np.arange(
            params.interval_min,
            params.interval_max + params.interval_spacing,  # type: ignore
            params.interval_spacing,
        ).tolist()
    else:
        interval_contours = []

    if params.fixed_contours != "" and params.fixed_contours is not None:
        if isinstance(params.fixed_contours, str):
            fixed_contours = re.split(",", params.fixed_contours.replace(" ", ""))
            fixed_contours = [float(c) for c in fixed_contours]
        elif isinstance(params.fixed_contours, float):
            fixed_contours = [params.fixed_contours]
    else:
        fixed_contours = []  # type: ignore

    contours = np.unique(np.sort(interval_contours + fixed_contours)).tolist()
    return contours


def find_curves(
    vertices: np.ndarray,
    parts: np.ndarray,
    params: TrendLineDetectionParameters | None = None,
) -> list[list[list[float]]]:
    """
    Find curves in a set of points.

    :param vertices: Vertices for points.
    :param parts: Identifier for points belong to common parts.
    :param params: Trend line detection parameters.

    :return: List of curves.
    """
    if params is None:
        params = TrendLineDetectionParameters()

    tri = Delaunay(vertices, qhull_options="QJ")
    if tri.simplices is None:  # pylint: disable=no-member
        return []

    simplices: np.ndarray = tri.simplices  # pylint: disable=no-member

    edges = np.vstack(
        (
            simplices[:, :2],
            simplices[:, 1:],
            simplices[:, ::2],
        )
    )
    edges = np.sort(edges, axis=1)
    edges = np.unique(edges, axis=0)
    distances = np.linalg.norm(vertices[edges[:, 0]] - vertices[edges[:, 1]], axis=1)
    distance_sort = np.argsort(distances)
    edges, distances = edges[distance_sort, :], distances[distance_sort]

    if params.max_distance is not None:
        edges = edges[distances <= params.max_distance, :]

    # Reject edges with same vertices id
    edge_parts = parts[edges]
    edges = edges[edge_parts[:, 0] != edge_parts[:, 1]]

    if params.azimuth is not None and params.azimuth_tol is not None:
        ind = filter_segments_orientation(
            vertices, edges, params.azimuth, params.azimuth_tol
        )
        edges = edges[ind]

    # Walk edges until no more edges can be added
    mask = np.ones(vertices.shape[0], dtype=bool)
    out_curves = []

    for ind in range(edges.shape[0]):
        if not np.any(mask[edges[ind]]):
            continue

        mask[edges[ind]] = False
        path = [edges[ind]]
        path, mask = walk_edges(
            path, edges[ind], edges, vertices, params.damping, mask=mask
        )
        path, mask = walk_edges(
            path, edges[ind][::-1], edges, vertices, params.damping, mask=mask
        )
        if len(path) < params.min_edges:
            continue

        out_curves.append(path)

    return out_curves


def walk_edges(  # pylint: disable=too-many-arguments
    path: list,
    incoming: list,
    edges: np.ndarray,
    vertices: np.ndarray,
    damping: float = 0.0,
    *,
    mask: np.ndarray | None = None,
) -> tuple[list, np.ndarray]:
    """
    Find all edges connected to a point.

    :param path: Current list of edges forming a path.
    :param incoming: Incoming edge.
    :param edges: All edges.
    :param vertices: Direction of the edges.
    :param damping: Damping factor between [0, 1] for the path roughness.
    :param mask: Mask for nodes that have already been visited.

    :return: Edges connected to point.
    """
    if mask is None:
        mask = np.ones(edges.max() + 1, dtype=bool)
        mask[np.hstack(path).flatten()] = False

    if damping < 0 or damping > 1:
        raise ValueError("Damping must be between 0 and 1.")

    neighbours = np.where(
        np.any(edges == incoming[1], axis=1) & np.any(mask[edges], axis=1)
    )[0]

    if len(neighbours) == 0:
        return path, mask

    # Outgoing candidate nodes
    candidates = edges[neighbours][edges[neighbours] != incoming[1]]

    vectors = vertices[candidates, :] - vertices[incoming[1], :]
    in_vec = np.diff(vertices[incoming, :], axis=0).flatten()
    dot = np.dot(in_vec, vectors.T)

    if not np.any(dot > 0):
        return path, mask

    # Remove backward vectors
    vectors = vectors[dot > 0, :]
    candidates = candidates[dot > 0]
    dot = dot[dot > 0]

    # Compute the angle between the incoming vector and the outgoing vectors
    vec_lengths = np.linalg.norm(vectors, axis=1)
    angle = np.arccos(dot / (np.linalg.norm(in_vec) * vec_lengths) - 1e-10)

    # Minimize the torque
    sub_ind = np.argmin(angle ** (1 - damping) * vec_lengths)
    outgoing = [incoming[1], candidates[sub_ind]]
    mask[candidates[sub_ind]] = False
    path.append(outgoing)

    # Continue walking
    path, mask = walk_edges(path, outgoing, edges, vertices, damping, mask=mask)

    return path, mask


def filter_segments_orientation(
    vertices: np.ndarray, edges: np.ndarray, azimuth: float, azimuth_tol: float
):
    """
    Filter segments orientation.

    :param vertices: Vertices for points.
    :param edges: Edges for points.
    :param azimuth: Filter angle (degree) on segments orientation, clockwise from North.
    :param azimuth_tol: Tolerance (degree) on the azimuth.

    :return: Array of boolean.
    """
    vectors = vertices[edges[:, 1], :] - vertices[edges[:, 0], :]
    test_vector = np.array([np.sin(np.deg2rad(azimuth)), np.cos(np.deg2rad(azimuth))])

    angles = np.arccos(np.dot(vectors, test_vector) / np.linalg.norm(vectors, axis=1))

    return np.logical_or(
        np.abs(angles) < np.deg2rad(azimuth_tol),
        np.abs(angles - np.pi) < np.deg2rad(azimuth_tol),
    )



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2024-2025 Mira Geoscience Ltd.                                '
#                                                                              '
#  This file is part of curve-apps package.                                    '
#                                                                              '
#  curve-apps is distributed under the terms and conditions of the MIT License '
#  (see LICENSE file at the root of this source code package).                 '
#                                                                              '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

from __future__ import annotations

import logging
import tempfile
from abc import abstractmethod
from pathlib import Path

from geoapps_utils.driver.data import BaseData
from geoapps_utils.driver.driver import BaseDriver
from geoh5py.groups import UIJsonGroup
from geoh5py.objects import Curve, ObjectBase
from geoh5py.shared.utils import fetch_active_workspace
from geoh5py.ui_json import InputFile


logger = logging.getLogger(__name__)


class BaseCurveDriver(BaseDriver):
    """
    Driver for the edge detection application.

    :param parameters: Application parameters.
    """

    _parameter_class: type[BaseData]
    _default_name: str

    def __init__(self, parameters: BaseData | InputFile):
        self._out_group = None
        if isinstance(parameters, InputFile):
            parameters = self._parameter_class.build(parameters)

        # TODO need to re-type params in base class
        super().__init__(parameters)

    @property
    def out_group(self) -> UIJsonGroup | None:
        """Output container group."""

        if self._out_group is None:
            self._out_group = UIJsonGroup.create(
                workspace=self.workspace,
                name=self.params.title,
            )
            self._out_group.options = InputFile.stringify(  # type: ignore
                InputFile.demote(self.params.input_file.ui_json)
            )

        return self._out_group

    def store(self, curve: Curve):
        """
        Update container group and monitoring directory.

        :param curve: Curve object to store.
        """

        with fetch_active_workspace(self.workspace, mode="r+") as workspace:
            self.update_monitoring_directory(
                curve if self.out_group is None else self.out_group
            )
            logger.info(
                "Curve object '%s' saved to '%s'.",
                self.params.output.export_as,
                str(workspace.h5file),
            )

    @abstractmethod
    def make_curve(self):
        pass

    def run(self):
        """
        Run method of the driver.
        """
        logging.info("Begin Process ...")
        curve = self.make_curve()
        logging.info("Process Complete.")
        self.store(curve)

    @property
    def params(self) -> BaseData:
        """Application parameters."""
        return self._params

    @params.setter
    def params(self, val: BaseData):
        if not isinstance(val, BaseData):
            raise TypeError("Parameters must be a BaseData subclass.")
        self._params = val

    def add_ui_json(self, entity: ObjectBase | UIJsonGroup) -> None:
        """
        Add ui.json file to entity.

        :param entity: Object to add ui.json file to.
        """

        with tempfile.TemporaryDirectory() as temp_dir:
            filepath = Path(temp_dir) / f"{self.params.name}.ui.json"
            self.params.write_ui_json(filepath)

            entity.add_file(str(filepath))



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2024-2025 Mira Geoscience Ltd.                                '
#                                                                              '
#  This file is part of curve-apps package.                                    '
#                                                                              '
#  curve-apps is distributed under the terms and conditions of the MIT License '
#  (see LICENSE file at the root of this source code package).                 '
#                                                                              '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

from __future__ import annotations

from pathlib import Path
from typing import ClassVar

from geoapps_utils.driver.data import BaseData
from geoh5py.data import FloatData
from geoh5py.groups import UIJsonGroup
from geoh5py.objects import Grid2D
from pydantic import BaseModel, ConfigDict

from curve_apps import assets_path


class EdgeSourceParameters(BaseModel):
    """
    Source parameters expected by the ui.json file format.

    :param objects: A Grid2D source object.
    :param data: Data values to find edges on.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    objects: Grid2D
    data: FloatData


class EdgeDetectionParameters(BaseModel):
    """
    Edge detection parameters.

    :param line_length: Minimum accepted pixel length of detected lines. (Hough)
    :param line_gap: Maximum gap between pixels to still form a line. (Hough)
    :param sigma: Standard deviation of the Gaussian filter. (Canny)
    :param threshold: Value threshold. (Hough)
    :param window_size: Size of the window to search for lines.
    :param merge_length: Minimum length between nodes that should be merged.
    """

    line_length: int = 1
    line_gap: int = 1
    sigma: float = 10
    threshold: int = 1
    window_size: int | None = None
    merge_length: float | None = None


class EdgeOutputParameters(BaseModel):
    """
    Output parameters.

    :param export_as: Name of the output entity.
    :param out_group: Name of the output group.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    export_as: str | None = "edges"
    out_group: UIJsonGroup | None = None


class EdgeParameters(BaseData):
    """
    Edge detection parameters for use with `edges.driver`.

    :param detection: Detection parameters expected for the edge detection.
    :param source: Parameters for the source object and data.
    :param output: Output parameters.
    """

    name: ClassVar[str] = "edges"
    default_ui_json: ClassVar[Path] = assets_path() / "uijson/edges.ui.json"
    title: ClassVar[str] = "Edge Detection"
    run_command: ClassVar[str] = "curve_apps.edges.driver"

    conda_environment: str = "curve_apps"
    source: EdgeSourceParameters
    detection: EdgeDetectionParameters
    output: EdgeOutputParameters = EdgeOutputParameters()



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2024-2025 Mira Geoscience Ltd.                                '
#                                                                              '
#  This file is part of curve-apps package.                                    '
#                                                                              '
#  curve-apps is distributed under the terms and conditions of the MIT License '
#  (see LICENSE file at the root of this source code package).                 '
#                                                                              '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

from __future__ import annotations

import logging
import sys

import numpy as np
from geoapps_utils.utils.formatters import string_name
from geoapps_utils.utils.transformations import rotate_xyz
from geoh5py.objects import Curve, Grid2D
from geoh5py.ui_json import InputFile, utils
from skimage import measure

from curve_apps.contours.params import ContourParameters
from curve_apps.driver import BaseCurveDriver
from curve_apps.utils import (
    image_to_grid_coordinate_transfer,
    interp_to_grid,
    set_vertices_height,
)


logger = logging.getLogger(__name__)


class ContoursDriver(BaseCurveDriver):
    """
    Driver for the detection of contours within geoh5py objects.

    :param parameters: Application parameters.
    """

    _parameter_class = ContourParameters

    def __init__(self, parameters: ContourParameters | InputFile):
        super().__init__(parameters)

    def make_curve(self):
        """
        Make curve object from contours detected in source data.
        """

        with utils.fetch_active_workspace(self.workspace, mode="r+"):
            logger.info("Generating contours ...")

            entity = self.params.source.objects
            data = self.params.source.data

            if isinstance(self.params.source.objects, Grid2D):
                x_grid = entity.origin["x"] + (
                    entity.u_cell_size * np.arange(entity.shape[0])
                )
                y_grid = entity.origin["y"] + (
                    entity.v_cell_size * np.arange(entity.shape[1])
                )
                grid = [x_grid, y_grid]
                data = data.values.reshape(entity.shape[::-1], order="C")

            else:
                grid, data = interp_to_grid(
                    self.params.source.objects,
                    self.params.source.data.values,
                    self.params.detection.resolution,
                    self.params.detection.max_distance,
                )

            locations, edges, values = ContoursDriver.get_contours(
                grid, data, self.params.detection.contours
            )

            if isinstance(entity, Grid2D):
                locations = rotate_xyz(
                    locations, center=entity.origin.tolist(), theta=entity.rotation
                )

            if self.params.output.z_value:
                locations = np.c_[locations, values]
            else:
                locations = set_vertices_height(locations, self.params.source.objects)

            curve = Curve.create(
                self.workspace,
                name=string_name(self.params.output.export_as),
                vertices=locations,
                cells=edges,
                parent=self.out_group,
            )
            curve.add_data(
                {
                    self.params.source.data.name: {
                        "association": "VERTEX",
                        "values": values,
                    }
                }
            )

            return curve

    @staticmethod
    def get_contours(
        grid: list[np.ndarray], data: np.ndarray, contour_list: list[float]
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Return vertices, edges, and values for contours.

        :param grid: list of x and y grids.
        :param data: 2D array of data living in grid.
        :param contour_list: list of contour values.
        """

        interp = image_to_grid_coordinate_transfer(data, grid)
        vertices, edges, values = [], [], []
        v0 = 0
        for contour in contour_list:
            coords = measure.find_contours(data, contour)
            if not coords:
                continue

            for coord in coords:
                segment_vertices = interp(coord[:, 1], coord[:, 0])
                nv = len(segment_vertices)
                segment_edges = np.c_[np.arange(nv - 1), np.arange(1, nv)] + v0
                vertices.append(segment_vertices)
                edges.append(segment_edges)
                values.append([contour] * nv)
                v0 += nv

        if not vertices:
            raise ValueError(
                "No contours detected. Check that the requested contour "
                "values are within the bounds of the data."
            )

        return (
            np.vstack(vertices),
            np.vstack(edges).astype("uint32"),
            np.hstack(values),
        )


if __name__ == "__main__":
    file = sys.argv[1]
    ifile = InputFile.read_ui_json(file)

    driver = ContoursDriver(ifile)
    driver.run()
